#include "comm.h"
#include "sensors.h"
#include "pumps.h"
#include "pumpcontroller.h"

CCommunicator::CCommunicator() {
	//
}

CCommunicator::~CCommunicator() {
	Stop();
}

bool CCommunicator::Start(size_t nParam) {
	try {
		// set up the websocketpp - be silend on all channels except failure
		mServer.clear_access_channels(websocketpp::log::alevel::all);
		mServer.set_access_channels(websocketpp::log::alevel::fail);

		mServer.init_asio();
		mServer.set_reuse_addr(true);
		mServer.set_message_handler(websocketpp::lib::bind(&CCommunicator::On_Message, this, websocketpp::lib::placeholders::_1, websocketpp::lib::placeholders::_2));

		// start the communications thread
		mComm_Thread = std::make_unique<std::thread>(&CCommunicator::Thread_Fnc, this);
	}
	catch (websocketpp::exception const& e) {
		std::cerr << e.what() << std::endl;
		return false;
	}
	catch (std::exception const& e) {
		std::cerr << "Unhandled exception: " << e.what() << std::endl;
		return false;
	}

	return true;
}

bool CCommunicator::Stop() {
	if (!mServer.stopped())
		mServer.stop();

	if (mComm_Thread && mComm_Thread->joinable())
		mComm_Thread->join();

	return true;
}

void CCommunicator::On_Message(websocketpp::connection_hdl hdl, websocketpp::server<websocketpp::config::asio>::message_ptr msg) {
	std::cout << "Message received: " << msg->get_payload() << std::endl;

	try
	{
		// extract the payload
		std::string payload = msg->get_payload();

		// sensor reading, e.g. read-sensor=12, where 12 is a past timestamp (the server may not respond, if the timestamp is newer or equal)
		if (payload.substr(0, 11) == "read-sensor" && payload.find('=') != std::string::npos && payload.length() > 12) {

			// extract timestamp
			auto ts = std::stoull(payload.substr(12));
			auto curTs = Svc.Get<CSensor, 0>().Get_Timestamp();
			// compare timestamps, respond if none or older
			if (ts <= 0 || ts < curTs) {
				auto valStr = "sensor-value=" + std::to_string(Svc.Get<CSensor, 0>().Get_Sensor_Value()) + "=" + std::to_string(curTs);
				mServer.send(hdl, valStr, msg->get_opcode());
			}
		}
		// pump basal reading, responds with basal rate
		else if (payload == "read-pump-basal") {
			auto valStr = "pump-basal=" + std::to_string(Svc.Get<CPump_Controller>().Get_Basal_Rate());
			mServer.send(hdl, valStr, msg->get_opcode());
		}
		// water pump params reading, responds with comma-delimited pumping rates
		else if (payload == "read-wp-params") {
			auto valStr = "wp=" + std::to_string(Svc.Get<CPump, 0>().Get_Current_Rate()) + "," + std::to_string(Svc.Get<CPump, 1>().Get_Current_Rate());
			mServer.send(hdl, valStr, msg->get_opcode());
		}
		// eat carbs, no response
		else if (payload.starts_with("eat-carbs") && payload.find('=') != std::string::npos) {

			auto carbs = std::stod(payload.substr(10));
			if (carbs > 0 && carbs < 150) {
				Svc.Get<CPump_Controller>().Eat_Carbs(carbs);
			}
			else {
				mServer.send(hdl, "error=invalid-carbs", msg->get_opcode());
			}
		}
		// dose bolus, no response
		else if (payload.starts_with("dose-bolus") && payload.find('=') != std::string::npos) {

			auto bolus = std::stod(payload.substr(11));
			if (bolus > 0 && bolus < 20) {
				Svc.Get<CPump_Controller>().Dose_Bolus(bolus);
			}
			else {
				mServer.send(hdl, "error=invalid-bolus", msg->get_opcode());
			}
		}
		// set basal rate, no response
		else if (payload.starts_with("set-basal") && payload.find('=') != std::string::npos) {

			auto basal = std::stod(payload.substr(10));
			if (basal >= 0 && basal < 10) {
				Svc.Get<CPump_Controller>().Set_Basal_Rate(basal);
			}
			else {
				mServer.send(hdl, "error=invalid-basal", msg->get_opcode());
			}
		}
	}
	catch (websocketpp::exception const& e) {
		std::cout << "Communication failed because: " << e.what() << std::endl;
	}
}

void CCommunicator::Thread_Fnc() {
	try {
		mServer.listen(9002);
		mServer.start_accept();
		mServer.run();
	}
	catch (websocketpp::exception const& e) {
		std::cout << e.what() << std::endl;
	}
	catch (std::exception const& e) {
		std::cout << "Unhandled exception: " << e.what() << std::endl;
	}
}
