#include <iostream>
#include <csignal>
#include <condition_variable>
#include <mutex>
#include "wiringPiBridge.h"

#include "service.h"
#include "comm.h"
#include "sensors.h"
#include "pumps.h"
#include "pumpcontroller.h"

CService_Dispatcher Svc;

namespace {
	bool gIs_Running = false;
	std::mutex gRunning_Mtx;
	std::condition_variable gRunning_Cv;
}

void sigint_handler(int signo) {
	std::unique_lock<std::mutex> lck;
	gIs_Running = false;
	gRunning_Cv.notify_all();
}

int main(int argc, char** argv) {

	// map termination (SIGINT) signal
	gIs_Running = true;
	signal(SIGINT, &sigint_handler);

	// setup wiringPi
	wiringPiSetup();

	// start all services; we do not handle dependencies on a structural level, so start up all services in a dependent order
	Svc.Start<CSensor, 0>();
	Svc.Start<CSensor, 1>();
	Svc.Start<CPump, 0>();
	Svc.Start<CPump, 1>();
	Svc.Start<CPump_Controller>();
	Svc.Start<CCommunicator>();

	// waiting scope - just sleep and wait for termination signal
	{
		std::unique_lock<std::mutex> lck(gRunning_Mtx);

		while (gIs_Running)
			gRunning_Cv.wait(lck);
	}

	// tear down all services; again, tear down in reverse order to ensure no dependencies exist between stopped and running services
	Svc.Stop<CCommunicator>();
	Svc.Stop<CPump_Controller>();
	Svc.Stop<CPump, 1>();
	Svc.Stop<CPump, 0>();
	Svc.Stop<CSensor, 1>();
	Svc.Stop<CSensor, 0>();

	return 0;
}
