#include "pumpcontroller.h"
#include "sensors.h"
#include "pumps.h"

#include <chrono>
#include "wiringPiBridge.h"

#undef min
#undef max

// shared variables

double gIns = 0;
double gCarb = 0;

std::mutex gSD_Mtx;
std::condition_variable gSD_cv;
bool gSD_Done = false;

/*
 * Terminal filter to retrieve calculated activity indexes
 */
class CTerminal_Filter : public virtual scgms::IFilter, public virtual refcnt::CNotReferenced {

	protected:
		double mAct_Insulin = 0;
		double mAct_Carbs = 0;

	public:
		CTerminal_Filter() = default;
		CTerminal_Filter(CTerminal_Filter&& other) noexcept = default;
		CTerminal_Filter& operator=(CTerminal_Filter&& other) noexcept = default;

		CTerminal_Filter(const CTerminal_Filter&) = delete;
		CTerminal_Filter& operator=(const CTerminal_Filter&) = delete;

		double Get_Insulin_Activity() const {
			return gIns;
		}

		double Get_Carbs_Activity() const {
			return gCarb;
		}

		virtual HRESULT IfaceCalling Configure(scgms::IFilter_Configuration* configuration, refcnt::wstr_list* error_description) {
			return S_OK;
		}

		virtual HRESULT IfaceCalling Execute(scgms::IDevice_Event* event) {

			// RAII scope
			{
				scgms::UDevice_Event evt{ event };

				if (evt.event_code() == scgms::NDevice_Event_Code::Level) {
					if (evt.signal_id() == scgms::signal_Virtual[0]) {
						mAct_Insulin = evt.level();
						gIns = evt.level();
					}
					else if (evt.signal_id() == scgms::signal_Virtual[1]) {
						mAct_Carbs = evt.level();
						gCarb = evt.level();
					}
				}
				else if (evt.event_code() == scgms::NDevice_Event_Code::Shut_Down) {
					std::unique_lock<std::mutex> lck(gSD_Mtx);
					gSD_Done = true;
					gSD_cv.notify_all();
				}
			}

			//event->Release();
			return S_OK;
		}
};




CPump_Controller::CPump_Controller() {
	//
}

CPump_Controller::~CPump_Controller() {
	//
}

void CPump_Controller::Ctl_Thread_Fnc() {

	while (mRunning) {

		std::unique_lock<std::mutex> lck(mRun_Mtx);

		// trigger sensors
		const double curIg = Svc.Get<CSensor, 0>().Trigger_And_Await_Measurement();
		const double curRemote = Svc.Get<CSensor, 1>().Trigger_And_Await_Measurement();

		if (mExecutor)
		{
			scgms::UDevice_Event evt{ scgms::NDevice_Event_Code::Level };
			evt.device_id() = Invalid_GUID;
			evt.signal_id() = scgms::signal_IG;
			evt.level() = curIg;
			evt.segment_id() = 1;
			evt.device_time() = Unix_Time_To_Rat_Time(mStart_Sim_Time + mCurrent_Sim_Tick * 5 * 60);

			mExecutor.Execute(std::move(evt));

			mCurrent_Sim_Tick++;
		}

		// desired steady level of IG
		constexpr double Steady_Ig = 15.0;

		double iaAdjust = 0;
		double caAdjust = 0;

		if (curIg < Steady_Ig) {
			caAdjust += ((Steady_Ig - curIg) / Steady_Ig) * 0.9; // TODO: empirically tune this constant
		}
		else {
			iaAdjust += ((curIg - Steady_Ig) / Steady_Ig) * 0.9; // TODO: empirically tune this constant
		}

		// IA = max at 0.07 (2.0 U/hr + 4 U bolus) --> Pump 0 at 1024
		// CA = max at 0.6 (60 g CHO) --> Pump 1 at 1024

		constexpr double iaFactorMax = 0.06; //orig = 0.07;
		constexpr double caFactorMax = 0.4;  //orig = 0.6;

		const double iaFactor = std::clamp(iaAdjust + (mTerminal_Filter->Get_Insulin_Activity() / iaFactorMax), 0.0, 1.0);
		const double caFactor = std::clamp(caAdjust + (mTerminal_Filter->Get_Carbs_Activity() / caFactorMax), 0.0, 1.0);

		// low IG, would further decrease
		if (curIg < 2.1 && caFactor < iaFactor) {
			Svc.Get<CPump, 0>().Set_Current_Rate(0);
			Svc.Get<CPump, 1>().Set_Current_Rate(0);
		}
		// high IG, would further increase
		else if (curIg > 18.0 && caFactor > iaFactor) {
			Svc.Get<CPump, 0>().Set_Current_Rate(0);
			Svc.Get<CPump, 1>().Set_Current_Rate(0);
		}
		else {
			constexpr int effectiveOffset = 180;

			Svc.Get<CPump, 1>().Set_Current_Rate(effectiveOffset + static_cast<int>(iaFactor * (1024 - effectiveOffset)));
			Svc.Get<CPump, 0>().Set_Current_Rate(effectiveOffset + static_cast<int>(caFactor * (1024 - effectiveOffset)));
		}

		mRun_Cv.wait_for(lck, std::chrono::milliseconds(500));
	}

	// turn off both pumps
	Svc.Get<CPump, 0>().Set_Current_Rate(0);
	Svc.Get<CPump, 1>().Set_Current_Rate(0);
}

void CPump_Controller::SCGMS_Thread_Fnc() {

	// start the executor
	refcnt::Swstr_list errors;
	mExecutor = scgms::SFilter_Executor{ mConfiguration.get(), nullptr, nullptr, errors, mTerminal_Filter.get() };

	errors.for_each([](auto str) { std::wcerr << str << std::endl;	});

	if (!mExecutor) {
		std::wcerr << L"Could not execute the filters!" << std::endl;
		return;
	}

	// wait for shutdown
	{
		std::unique_lock<std::mutex> lck(gSD_Mtx);

		while (!gSD_Done)
			gSD_cv.wait(lck);
	}
}

double CPump_Controller::Get_Basal_Rate() const {
	return mCurrent_Basal_Rate;
}

void CPump_Controller::Set_Basal_Rate(double rate) {
	mCurrent_Basal_Rate = rate;

	Send_Level_Event(rate, scgms::signal_Requested_Insulin_Basal_Rate);
}

void CPump_Controller::Dose_Bolus(double amount) {
	Send_Level_Event(amount, scgms::signal_Requested_Insulin_Bolus);
}

void CPump_Controller::Eat_Carbs(double amount) {
	Send_Level_Event(amount, scgms::signal_Carb_Intake);
}

void CPump_Controller::Send_Level_Event(double level, const GUID& signal_id) {
	
	std::unique_lock<std::mutex> lck(mRun_Mtx);

	scgms::UDevice_Event evt{ scgms::NDevice_Event_Code::Level };
	evt.device_id() = Invalid_GUID;
	evt.signal_id() = signal_id;
	evt.level() = level;
	evt.segment_id() = 1;
	evt.device_time() = Unix_Time_To_Rat_Time(mStart_Sim_Time + mCurrent_Sim_Tick * 5 * 60);

	mExecutor.Execute(std::move(evt));
}

bool CPump_Controller::Start(size_t nParam) {

	mRunning = true;
	mCurrent_Basal_Rate = 0.0;
	mStart_Sim_Time = time(nullptr);
	mCurrent_Sim_Tick = 0;

	refcnt::Swstr_list errors;

	HRESULT rc = E_FAIL;
	if (mConfiguration) {
		rc = mConfiguration->Load_From_File((Get_Application_Dir() / L"prva.ini").wstring().c_str(), errors.get());
	}

	errors.for_each([](auto str) {
		std::wcerr << str << std::endl;
	});

	mTerminal_Filter = std::make_unique<CTerminal_Filter>();
	mTerminal_Filter->AddRef();

	mCtl_Thread = std::make_unique<std::thread>(&CPump_Controller::Ctl_Thread_Fnc, this);
	mScgms_Thread = std::make_unique<std::thread>(&CPump_Controller::SCGMS_Thread_Fnc, this);

	return Succeeded(rc);
}

bool CPump_Controller::Stop() {

	// evt scope
	{
		std::unique_lock<std::mutex> lck(mRun_Mtx);

		scgms::UDevice_Event evt{ scgms::NDevice_Event_Code::Shut_Down };
		evt.device_id() = Invalid_GUID;
		evt.signal_id() = scgms::signal_IG;
		evt.segment_id() = 1;
		evt.device_time() = Unix_Time_To_Rat_Time(mStart_Sim_Time + mCurrent_Sim_Tick * 5 * 60);

		mExecutor.Execute(std::move(evt));
	}

	// lock scope
	{
		std::unique_lock<std::mutex> lck(mRun_Mtx);

		mRunning = false;
		mRun_Cv.notify_all();
	}

	if (mCtl_Thread && mCtl_Thread->joinable())
		mCtl_Thread->join();

	if (mScgms_Thread && mScgms_Thread->joinable())
		mScgms_Thread->join();

	return true;
}
