#pragma once

#include <memory>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <ctime>

#include <rtl/FilterLib.h>
#include <rtl/rattime.h>

#include "service.h"

class CTerminal_Filter;

/*
 * Service class for pump controller and for SmartCGMS communications
 */
class CPump_Controller final : public IService
{
	private:
		// control thread
		std::unique_ptr<std::thread> mCtl_Thread;
		// SmartCGMS execution thread
		std::unique_ptr<std::thread> mScgms_Thread;

		// shared mutex
		std::mutex mRun_Mtx;
		// shared CV
		std::condition_variable mRun_Cv;

		// is running?
		bool mRunning = false;
		// current basal rate (U/hr, not the PWM stride)
		double mCurrent_Basal_Rate = 0;

		// stored SmartCGMS configuration
		scgms::SPersistent_Filter_Chain_Configuration mConfiguration;
		// SmartCGMS executor
		scgms::SFilter_Executor mExecutor;

		// stored terminal filter (we manage the lifetime)
		std::unique_ptr<CTerminal_Filter> mTerminal_Filter;

		// time of simulation start
		time_t mStart_Sim_Time = 0;
		// current simulation tick (current time = start + tick * 5min)
		time_t mCurrent_Sim_Tick = 0;

	protected:
		// control thread method
		void Ctl_Thread_Fnc();
		// SmartCGMS execution thread method
		void SCGMS_Thread_Fnc();

		// sends a level event to the SmartCGMS chain
		void Send_Level_Event(double level, const GUID& signal_id);

	public:
		CPump_Controller();
		~CPump_Controller();

		// retrieves current basal infusion rate
		double Get_Basal_Rate() const;
		// sets current basal infusion rate
		void Set_Basal_Rate(double rate);
		// doses a given bolus amount
		void Dose_Bolus(double amount);
		// ingests a given carbs amount
		void Eat_Carbs(double amount);

		// IService interface
		virtual bool Start(size_t nParam) override;
		virtual bool Stop() override;

		// INamed_Service concept
		constexpr static const char* Service_Name = "Pump controller";
};

