#include "pumps.h"

#include <chrono>
#include "wiringPiBridge.h"

CPump::CPump() {
	//
}

CPump::~CPump() {
	//
}

void CPump::Thread_Fnc() {

	int realRate = 0;

	while (mRunning) {

		std::unique_lock<std::mutex> lck(mRun_Mtx);

		// periodically check for changes
		mRun_Cv.wait_for(lck, std::chrono::milliseconds(2500));

		// stepping and threshold for stepping
		constexpr int threshold = 5;

		// incrementally change the rate, so the pump does not go from zero to full instantly
		// the sudden change of pump rate moves the pump mechanically in the container, so we rather sacrifice a small amount of time to start up the pump
		// this also eliminates a part of sensor noise due to "waves" created by the pump moving in the tank
		while (realRate != mCurrent_Rate && mRunning) {
			int delta = mCurrent_Rate - realRate;
			// under threshold - set the rate to the final value
			if (std::abs(delta) < threshold) {
				realRate = mCurrent_Rate;
			}
			// should decrease? decrease in steps
			else if (delta < 0) {
				realRate -= threshold;
			}
			// should increase? increase in steps
			else {
				realRate += threshold;
			}

			// update the PWM
			pwmWrite(mPwm_Pin, realRate);

			// wait a small amount of time before next stepping
			mRun_Cv.wait_for(lck, std::chrono::milliseconds(50));
		}
	}

}

int CPump::Get_Current_Rate() const {
	return mCurrent_Rate;
}

void CPump::Set_Current_Rate(int rate) {

	std::unique_lock<std::mutex> lck(mRun_Mtx);

	mCurrent_Rate = std::min(1024, std::max(0, rate)); // limit to 0-1024

	// notify the pump thread to gradually change the real PWM stride
	mRun_Cv.notify_all();
}

bool CPump::Start(size_t nParam) {

	mPump_Id = nParam;
	mRunning = true;
	mCurrent_Rate = 0;

	if (nParam == 0) {
		mPwm_Pin = 26; // GPIO 12, wiringPi 26
	}
	else if (nParam == 1) {
		mPwm_Pin = 23; // GPIO 13, wiringPi 23
	}
	else {
		std::cerr << "Unknown pump number " << nParam << std::endl;
		return false;
	}

	pinMode(mPwm_Pin, PinMode(PinModes::_PWM_OUTPUT));
	// clock divider 16
	pwmSetClock(16);
	// mark:set mode
	pwmSetMode(PWM_MODE_MS);

	mPump_Thread = std::make_unique<std::thread>(&CPump::Thread_Fnc, this);

	return true;
}

bool CPump::Stop() {

	// lock scope
	{
		std::unique_lock<std::mutex> lck(mRun_Mtx);

		mRunning = false;
		mRun_Cv.notify_all();
	}

	if (mPump_Thread && mPump_Thread->joinable())
		mPump_Thread->join();

	pwmWrite(mPwm_Pin, 0);

	return true;
}
