#include "sensors.h"

#include <chrono>
#include <array>
#include "wiringPiBridge.h"

#ifndef _WIN32
#include <pthread.h>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <unistd.h>
#include <sys/resource.h>

// set FIFO priority for RT scheduler
void setFIFOPriority(auto pthr) {
	int policy, s;
	struct sched_param param;

	s = pthread_getschedparam(pthr, &policy, &param);
	if (s != 0) {
		std::cerr << "Unable to obtain scheduler parameters" << std::endl;
		return;
	}

	policy = SCHED_FIFO;
	param.__sched_priority = sched_get_priority_min(SCHED_FIFO);
	s = pthread_setschedparam(pthr, policy, &param);
	if (s != 0) {
		std::cerr << "Unable to set FIFO RT policy" << std::endl;
		return;
	}
}
#else
void setFIFOPriority(auto pthr) {
	//
}
#endif

CSensor::CSensor() {
	//
}

CSensor::~CSensor() {
	//
}

void activeWait(std::chrono::steady_clock::duration dur) {
	auto tp_start = std::chrono::high_resolution_clock::now();
	while (std::chrono::high_resolution_clock::now() - tp_start < dur)
		;
}

void CSensor::Thread_Fnc() {

	while (mRunning) {

		std::unique_lock<std::mutex> lck(mRun_Mtx);

		mRun_Cv.wait_for(lck, std::chrono::milliseconds(1800));

		if (!mMeasurement_Triggered || !mRunning)
			continue;

		mMeasurement_Triggered = false;

		double avgDiff = 0;
		constexpr size_t measureCount = 7;
		constexpr size_t medianIndex = measureCount / 2;
		std::array<double, measureCount> measurements;

		auto maxDiff = std::chrono::milliseconds(1000);

		bool isStable = false;

		while (!isStable) {

			for (size_t i = 0; i < measureCount; i++) {

				// it is crucial to hold the CPU for the entitery of time
				// if the scheduler passes the time to another process, it creates non-trivial delays in sensor responses
				// the timing must be precise (with a resolution of microseconds), otherwise measurements are not precise, as they are based on detecting the time of pin level change

				if (i != 0) {
					activeWait(std::chrono::milliseconds(50));
				}

				// send a 10uS pulse to the trigger pin
				digitalWrite(mTrig_Pin, PinLevel(PinLevels::_HIGH));
				activeWait(std::chrono::microseconds(10));
				digitalWrite(mTrig_Pin, PinLevel(PinLevels::_LOW));

				// if we mock the state, mock a 600uS delay
				mock::pinStateFor(mEcho_Pin, PinLevels::_HIGH, PinLevels::_LOW, std::chrono::microseconds(600));

				// detect echo start
				auto tp_start = std::chrono::high_resolution_clock::now();
				const auto fixed_start = tp_start;
				while (digitalRead(mEcho_Pin) == PinLevel(PinLevels::_LOW) || (tp_start - fixed_start > maxDiff))
					tp_start = std::chrono::high_resolution_clock::now();

				// detect echo end
				auto tp_end = std::chrono::high_resolution_clock::now();
				while (digitalRead(mEcho_Pin) == PinLevel(PinLevels::_HIGH) || (tp_end - tp_start > maxDiff))
					tp_end = std::chrono::high_resolution_clock::now();

				// calculate the difference in microseconds
				auto diff = static_cast<double>(std::chrono::duration_cast<std::chrono::microseconds>(tp_end - tp_start).count());

				// anything less than 100 is a failed reading (as there is no way to achieve such time without misreading an old ultrasound echo)
				if (diff < 100) {
					// invalid attempt, repeat
					i--;
					continue;
				}

				// add measurement and accumulate average
				measurements[i] = diff;
				avgDiff += diff;
			}

			// calculate the average and sort the values to obtain median
			avgDiff /= static_cast<double>(measureCount);
			std::sort(measurements.begin(), measurements.end());

			// median and mean must not have a difference greater than 100
			if (std::fabs(measurements[2] - avgDiff) > 100)
				isStable = false;
			else
			{
				// TODO: more checks?
				isStable = true;
			}
		}

		// approx 900 = max dist --> 2.0  mmol/L
		// approx 300 = min dist --> 20.0 mmol/L

		constexpr double Max_Dist = 900.0;
		constexpr double Min_Dist = 300.0;

		avgDiff = std::min(Max_Dist, std::max(avgDiff, Min_Dist));

		// convert time difference to concentration
		mCurrent_Value = 20.0 - ((avgDiff - Min_Dist)/(Max_Dist - Min_Dist)) * (20.0 - 2.0);

		mCurrent_Timestamp++;

		// if somebody expects the value, notify him
		mValue_Await_Cv.notify_one();
	}

}

void CSensor::Trigger_Measurement() {
	std::unique_lock<std::mutex> lck(mRun_Mtx);

	// just trigger the measurement and leave
	mMeasurement_Triggered = true;
	mRun_Cv.notify_all();
}

double CSensor::Trigger_And_Await_Measurement() {

	std::unique_lock<std::mutex> lck(mRun_Mtx);

	auto tsBefore = mCurrent_Timestamp;

	// trigger the measurement and wait for new reading
	mMeasurement_Triggered = true;
	mRun_Cv.notify_all();

	// wait while no newer value available (spurious wakeup)
	while (mCurrent_Timestamp == tsBefore)
		mValue_Await_Cv.wait(lck);

	return mCurrent_Value;
}

double CSensor::Get_Sensor_Value() const {
	return mCurrent_Value;
}

size_t CSensor::Get_Timestamp() const {
	return mCurrent_Timestamp;
}

bool CSensor::Start(size_t nParam) {

	mSensor_Id = nParam;
	mRunning = true;

	if (nParam == 0) {
		mTrig_Pin = 15; // GPIO 14, wiringPi 15
		mEcho_Pin = 16; // GPIO 15, wiringPi 16
	}
	else if (nParam == 1) {
		mTrig_Pin = 4;  // GPIO 23, wiringPi 4
		mEcho_Pin = 5;  // GPIO 24, wiringPi 5
	}
	else {
		std::cerr << "Unknown sensor index: " << nParam << std::endl;
		return false;
	}

	// set up pins - trigger pin in output mode, echo pin as input
	pinMode(mTrig_Pin, PinMode(PinModes::_OUTPUT));
	pinMode(mEcho_Pin, PinMode(PinModes::_INPUT));

	// ensure the trigger pin is pulled low
	digitalWrite(mTrig_Pin, PinLevel(PinLevels::_LOW));

	// start the thread
	mSensor_Thread = std::make_unique<std::thread>(&CSensor::Thread_Fnc, this);

	// set the FIFO RT priority, to hold CPU during the measurements
	setFIFOPriority(mSensor_Thread->native_handle());

	return true;
}

bool CSensor::Stop() {

	// lock scope
	{
		std::unique_lock<std::mutex> lck(mRun_Mtx);

		mCurrent_Timestamp++;
		mRunning = false;
		mRun_Cv.notify_all();
	}

	if (mSensor_Thread && mSensor_Thread->joinable())
		mSensor_Thread->join();

	return true;
}
