#pragma once

#include <unordered_map>
#include <tuple>
#include <typeindex>
#include <typeinfo>
#include <iostream>
#include <limits>

/*
 * Base service class
 */
class IService {
	public:
		virtual ~IService() = default;

		// starts the service with the given numeric parameter (usually an identifier, index or position, depending on service type and semantics)
		virtual bool Start(size_t nParam) = 0;
		// stops the service
		virtual bool Stop() = 0;
};

// concept for named service, so the dispatcher could inform the user about it during startup/teardown
template<typename T>
concept INamed_Service = requires(T t) {
	{ T::Service_Name } -> std::convertible_to<const char*>;
} && std::derived_from<T, IService>;

/*
 * Service dispatcher class - manages services and their lifetime
 */
class CService_Dispatcher final
{
	public:
		// numeric parameter of a service, that has not been given any by the user
		static constexpr size_t Unspecified_Svc_Index = std::numeric_limits<size_t>::max();

	private:
		// key type for service map
		using key_type = std::tuple<std::type_index, size_t>;

		// hasher for std::unordered_map of services
		struct key_hash
		{
			std::size_t operator()(const key_type& k) const
			{
				std::hash<size_t> sthash;
				return std::get<0>(k).hash_code() ^ sthash(std::get<1>(k));
			}
		};

		// comparator for std::unordered_map of services
		struct key_equal {
			bool operator()(const key_type& v0, const key_type& v1) const
			{
				return (std::get<0>(v0) == std::get<0>(v1) && std::get<1>(v0) == std::get<1>(v1));
			}
		};

		// map of services, key is a tuple of type and numeric parameter, value is a pointer to the service instance
		std::unordered_map<std::tuple<std::type_index, size_t>, std::unique_ptr<IService>, key_hash, key_equal> mServices;

	public:
		CService_Dispatcher() = default;

		// starts a service with a given numerical parameter, if any; does not start the service twice with the same parameters
		template<INamed_Service TService, size_t NParam = Unspecified_Svc_Index>
		bool Start() {
			// create key to the map
			auto key = std::make_tuple<std::type_index, size_t>(typeid(TService), NParam);

			if (mServices.find(key) != mServices.end())
				return false;

			// insert a new service
			mServices.insert(std::make_pair(key, std::make_unique<TService>()));

			std::cout << "Starting service " << TService::Service_Name << "..." << std::endl;

			// call the start method to set up the service for operation
			bool success = mServices[key]->Start(NParam);

			std::cout << "Service " << TService::Service_Name << " started " << (success ? "successfully" : "unsuccessfully") << std::endl;

			return success;
		}

		// stops the service with a given numerical parameter
		template<INamed_Service TService, size_t NParam = Unspecified_Svc_Index>
		bool Stop() {
			// create key to the map
			auto key = std::make_tuple<std::type_index, size_t>(typeid(TService), NParam);

			if (mServices.find(key) == mServices.end())
				return false;

			std::cout << "Stopping service " << TService::Service_Name << "..." << std::endl;

			// stop the service operation - this should effectively stop all its threads and deallocate all resources except the RAII ones, that gets deallocated automatically below
			bool success = mServices[key]->Stop();

			std::cout << "Service " << TService::Service_Name << " stopped " << (success ? "successfully" : "unsuccessfully") << std::endl;

			// erases the service from the service map, effectively deallocating all RAII-based resources
			mServices.erase(key);

			return success;
		}

		// retrieves a service by a given type+parameter; throws an exception when no such service is found
		template<INamed_Service TService, size_t NParam = Unspecified_Svc_Index>
		TService& Get() {
			auto key = std::make_tuple<std::type_index, size_t>(typeid(TService), NParam);
			return static_cast<TService&>(*mServices.at(key)); // static_cast, as we are sure the type is correct
		}
};

// a global instance of service dispatcher
extern CService_Dispatcher Svc;
