#include "wiringPiBridge.h"

#if defined(MOCKED_PINS)

#include <array>

// total pin count (RPi4)
constexpr size_t PinCount = 32;

namespace {
	// mocked pin states
	std::array<::PinLevels, ::PinCount> PinStates;

	// pin future states (for timed mocking)
	struct PinFuture {
		// active mocking?
		bool active = false;
		// when to flip
		std::chrono::high_resolution_clock::time_point flipTime;
		// new level after flip
		::PinLevels futureLevel = ::PinLevels::_LOW;
	};

	// future states for timed mocking
	std::array<PinFuture, ::PinCount> PinFutures;
}

int wiringPiSetup() {

	// set up all pins to LOW state
	std::fill(PinStates.begin(), PinStates.end(), PinLevels::_LOW);
	// deactivate all futures
	for (auto& f : PinFutures) {
		f.active = false;
	}

	return 0;
}

void pinMode(int pinNo, int mode) {
	//
}

int digitalRead(int pinNo) {

	// verify timed futures and mocking
	auto& f = PinFutures.at(pinNo);
	if (f.active && std::chrono::high_resolution_clock::now() >= f.flipTime) {
		f.active = false;
		mock::pinState(pinNo, f.futureLevel);
	}

	return PinLevel(PinStates.at(pinNo));
}

void digitalWrite(int pinNo, int level) {
	PinStates.at(pinNo) = static_cast<PinLevels>(level);
}

void pwmWrite(int pinNo, int dutyCycle) {
	// not supported
}

void pwmSetClock(int) {
	// not supported
}

void pwmSetMode(int) {
	// not supported
}

namespace mock {

	// mock pin state
	void pinState(int pinNo, PinLevels level) {
		PinStates.at(pinNo) = static_cast<PinLevels>(level);
	}

	// mock pin state for certain amount of time
	void pinStateFor(int pinNo, PinLevels levelNow, PinLevels levelLater, std::chrono::high_resolution_clock::duration dur) {
		auto& f = PinFutures.at(pinNo);

		pinState(pinNo, levelNow);

		f.flipTime = std::chrono::high_resolution_clock::now() + dur;
		f.futureLevel = levelLater;
		f.active = true;
	}

}

#else

namespace mock {

	void pinState(int pinNo, int level) {
		// pins not mocked
	}

	void pinStateFor(int pinNo, PinLevels levelNow, PinLevels levelLater, std::chrono::high_resolution_clock::duration dur) {
		// pins not mocked
	}

}

#endif
